<?php 
//*************************************************************************************
//*************************************************************************************
// Utility functions for WordPress Rapid Secure Login
//*************************************************************************************
//*************************************************************************************

//*************************************************************************************
// Find the user associated with the given UUID
//*************************************************************************************
function rpsl_get_user_by_uuid( $uuid ) {
// Given a UUID, returns the user object or null if nothing found 
	global $wpdb;
	$uuid       = sanitize_text_field( $uuid );
	$table_name = $wpdb->prefix . "rpsl_devices";
	
	$user_id    = $wpdb->get_var( $wpdb->prepare(
						"SELECT user_id FROM " . $table_name . " WHERE rpsl_uuid = %s ", $uuid ));
	if ($user_id == 0) { 
		return null ;
	} else {
		return get_userdata($user_id);
	}
}

//*************************************************************************************
// Add a new device to a user account
//*************************************************************************************
function rpsl_add_device( $user_id, $rpsl_uuid, $device_name) {
// Adds a device to the user account 
	global $wpdb;
	
	$device_name = sanitize_text_field($device_name);
	$rpsl_uuid   = sanitize_text_field($rpsl_uuid);
	
	if(!is_int($user_id)) {
		rpsl_trace("rpsl_add_device: user id not int: $user_id");
		return null;
	}
	
	if(!empty($device_name)) {
		$device_name = substr($device_name, 0, 80);
	}
		
	$table_name = $wpdb->prefix . "rpsl_devices";
	$result     = $wpdb->insert($table_name, array(
													'rpsl_uuid'   =>$rpsl_uuid, 
													'user_id'   =>$user_id, 
													'devicename'=>$device_name, 
													'created'   =>current_time('mysql'),
													'last_used' =>current_time('mysql'),
													'status'    =>"ok",
												));
	return $result;
}
//*************************************************************************************
// Set the 'last used' value for the given device
//*************************************************************************************
function rpsl_set_device_last_used ($uuid) {
	global $wpdb;
	$uuid          = sanitize_text_field( $uuid );
	$table_name    = $wpdb->prefix . "rpsl_devices";

	$set_last_used = $wpdb->prepare(
						"UPDATE " . $table_name . " SET last_used = NOW() WHERE rpsl_uuid = %s ", array($uuid)	);
	$updates = $wpdb->query( $set_last_used );
	return;
}

function rpsl_delete_device($uuid){
	global $wpdb;
	$uuid       = sanitize_text_field( $uuid );
	$table_name = $wpdb->prefix . "rpsl_devices";
		
	$delete_query = $wpdb->prepare(
						"DELETE FROM " . $table_name . " WHERE rpsl_uuid = %s ",
						array($uuid)
					);
	$deletes = $wpdb->query( $delete_query );
}

function rpsl_delete_session_row($uuid) {
	global $wpdb;
	$uuid       = sanitize_text_field( $uuid );
	$table_name = $wpdb->prefix . "rpsl_sessions";
		
	$delete_query = $wpdb->prepare("DELETE FROM " . $table_name . " WHERE rpsession = %s ", $uuid);
	$deletes = $wpdb->query( $delete_query );
}


//*************************************************************************************
// Create new Database Entry for the service authentication certificate
//*************************************************************************************
function rpsl_import_sac_certificate($pfx_content, $password){
// This function takes the supplied pfx which is base64 encoded and extracts the certificate and private
// key into separate PEM files. The key file is protected with a password derived from
// the system-wide NONCE_KEY defined in config.php
// The new files are written to the <plugin>/certs folder as crt.pem and key.pem

	rpsl_trace("Entering rpsl_import_sac_certificate method");
	// Get some filenames for the server authentication cert and key
	
	$result = false;
	try{
		// Parse the PFX file
		$parse_ok = openssl_pkcs12_read( $pfx_content, $p12, $password );
		
		if(!$parse_ok) {
			// If parse unsuccessful, log the error
			rpsl_log_openssl_error();
			return false;
		}

		$sslconf = rpsl_cert_folder() . '/openssl.cnf';
		if(file_exists($sslconf)) {	
			$config_args = array('config' => $sslconf); 
		} else {
			$config_args = null;
		}

		$key_ok = openssl_pkey_export($p12['pkey'], $protected_key, rpsl_keyfile_password(), $config_args);

		if (!$key_ok) {
			$openssl_error = rpsl_log_openssl_error();
			if(rpsl_resolve_openssl_error($openssl_error)){
				clearstatcache();
				if(file_exists($sslconf)) {	$config_args = array('config' => $sslconf); }
				$key_ok = openssl_pkey_export($p12['pkey'], $protected_key, rpsl_keyfile_password(), $config_args);
			}
		}

		if ($parse_ok && $key_ok) {
			// Add Sac's public and private key to db
			$result = rpsl_db_add_sac_certificate($p12['cert'], $protected_key);
		} else {
			$openssl_error = rpsl_log_openssl_error();		
		}	
	} catch (Exception $e) {
		rpsl_trace("Exception: " . $e->getMessage());
	}
	
	rpsl_trace("Exiting rpsl_import_sac_certificate method: $result");
	
	return $result;
}

function rpsl_resolve_openssl_error($message) {
	
	if(strpos($message, 'error:02001002:system library:fopen:No such file or directory') !== false) {
		// if this error is detected, it is due to the server platform not having access to the default openssl_open cnf file.
		// move sample default conf into position and retry method.
		$sample_sslconf = rpsl_cert_folder() . '/sample-openssl.cnf';
		$sslconf = rpsl_cert_folder() . '/openssl.cnf';

		if(!file_exists($sample_sslconf)) {
			rpsl_trace("No sample-openssl.cnf file exists to attempt default openssl.cnf fix, unable to retry");
			return false;
		}

		rename($sample_sslconf, $sslconf);
		return true;
	}

	return false;
}


function rpsl_db_add_sac_certificate($public_key, $private_key) {
	
	global $wpdb;
	// Adds a sac to the db
	$table_name = $wpdb->prefix . "rpsl_certificates";
	$result     = $wpdb->insert($table_name, array(
												'created'   	=> current_time('mysql'), 
												'public_key'   	=> $public_key, 
												'private_key'	=> $private_key, 
												'type'			=> "SAC",
												'status'    	=> "ACTIVE",
												));

	if($result === false) {
		rpsl_trace("Unable to insert new sac certificate into the database.");
		return false;
	}
	else
	{
		// deactivate any other sac certs.
		// certificate rows which have status 1, are a SAC and are not the new row
		$new_row_id = $wpdb->insert_id;
		$update  = $wpdb->prepare("UPDATE " . $table_name . " SET status = 'DISABLED' where certificate_id != %d AND type = 'SAC' AND status = 'ACTIVE'", $new_row_id );
		$updated = $wpdb->query( $update );
		return true;
	}
}

//*************************************************************************************
// Create new Database Entry for the issuer certificate
//*************************************************************************************
function rpsl_add_issuer_certificate($public_key) {
	// Adds a sac to the db
	global $wpdb;
	
	if(empty($public_key))
	{
		rpsl_trace("rpsl_add_issuer_certificate: Certificate provided cannot be empty");
		return null;
	}

	try{
		$result = openssl_pkey_get_public($public_key);

		if($result === false){
			rpsl_trace("rpsl_add_issuer_certificate: Certificate not a valid X509 Public Key Certificate");
			return null;
		}
	} catch (Exception $e) {
		rpsl_trace("Exception: " . $e->getMessage());
		return null;
	}

	// check for duplicate issuer certificates.
	$table_name = $wpdb->prefix . "rpsl_certificates";
	$sql_duplicates = $wpdb->prepare("SELECT * FROM $table_name WHERE status = 'ACTIVE' AND type = 'TIC' AND public_key = %s", $public_key);
	$duplicates = $wpdb->get_row($sql_duplicates);

	if($duplicates !== null) {
		// Issuer Certificate already exists
		// don't add it but it is a successful state
		rpsl_trace("Issuer certificate with public key provided already exists. Skipping add.");
		return true;
	}

	$result     = $wpdb->insert($table_name, array(
							'created'   	=> current_time('mysql'), 
							'public_key'   	=> $public_key,
							'type'			=> "TIC", 
							'status'    	=> 'ACTIVE',
							));
	if($result === false) {
		rpsl_trace("Unable to insert issuer certificate into the database.");
	}

	return $result;
}

//*************************************************************************************
// Get the certificates available for this site
//*************************************************************************************
function rpsl_db_get_active_issuer_certificates() {
// Given a UUID, returns the user object or null if nothing found 
	global $wpdb;
	
	$table_name = $wpdb->prefix . "rpsl_certificates";
	
	$rows = $wpdb->get_results("SELECT * FROM " . $table_name . " WHERE status = 'ACTIVE' AND type = 'TIC' ");
	
	return $rows;
}

//*************************************************************************************
// Get the certificates available for this site
//*************************************************************************************
function rpsl_db_get_active_sac_certificate() {
// Given a UUID, returns the user object or null if nothing found 
	global $wpdb;
	
	$table_name = $wpdb->prefix . "rpsl_certificates";
	
	$first_active_sac_cert = $wpdb->get_row("SELECT * FROM " . $table_name . " WHERE status = 'ACTIVE' AND type = 'SAC'");
	
	return $first_active_sac_cert;
}

//*************************************************************************************
// Check the current status of RapID certificates
//*************************************************************************************
function rpsl_db_check_configuration(){

	$issuer_certificates = rpsl_db_get_active_issuer_certificates();
	$sac_certificate = rpsl_db_get_active_sac_certificate();
	
	$has_sac_cert = false;
	$has_issuer_cert = false;

	if(!empty($sac_certificate)){ $has_sac_cert = true; };
	if(!empty($issuer_certificates)) { $has_issuer_cert = true;}

	$result = array( 
		"CACert" 	=> $has_issuer_cert, 
		"RapidAuthCert" => $has_sac_cert, 
		"RapidAuthKey" 	=> $has_sac_cert
	);

    return $result;
}

//*************************************************************************************
// Get the password for the Rapid server authentication key file (unique per site)
//*************************************************************************************
function rpsl_keyfile_password() {

	global $wpdb;
	$table_name = $wpdb->prefix . "rpsl_config";
	$sql = "SELECT secretkey FROM $table_name ;";
	$secretkey = $wpdb->get_var( $sql );
	return $secretkey;
}

//*************************************************************************************
// Get the path to the certs folder and ensure it exists	
//*************************************************************************************
function rpsl_cert_folder() {
	$folder = plugin_dir_path( __FILE__ ) . "certs";
	wp_mkdir_p($folder); 
	return $folder;
}

//*************************************************************************************
// Get the path / url to the temp folder and ensure it exists
//*************************************************************************************
function rpsl_temp_folder() {
	$folder = plugin_dir_path( __FILE__ ) . "temp";
	wp_mkdir_p($folder); 
	return $folder;
}

//*************************************************************************************
// Clean up old files
//*************************************************************************************
function rpsl_delete_old_files($folder) {
  $files = glob($folder."/*");
  $now   = time();

  foreach ($files as $file)
    if (is_file($file))
      if ($now - filemtime($file) >= 1200) 
        unlink($file);

}

//*************************************************************************************
// Read / Write the Site code, name, login style etc
//*************************************************************************************
function rpsl_site_code() {
	// This is effectively the 'Platform' code
	if(substr(admin_url(), -9) == "wp-admin/") {
		return "W1"; // Standard location
	} else {
		return "W2"; // Non-standard location
	}
}

//-------------------------------------------------------------------------------------------------
function rpsl_set_site_name($name) {
	global $wpdb;
	$clean_name = sanitize_text_field( $name );
	$clean_name = str_replace(rpsl_delim(), " ", $clean_name); // Cannot have the QR code delimiter in the name
	$clean_name = str_replace("|", " ", $clean_name);  // .. or our Ajax delimiter
	
	
	$table_name = $wpdb->prefix . "rpsl_config";
	$update  = $wpdb->prepare("UPDATE " . $table_name . " SET site_name = %s ", $clean_name);
	$updated = $wpdb->query( $update );
	return;
}

function rpsl_site_name() {
	global $wpdb;
	$table_name = $wpdb->prefix . "rpsl_config";
	$sql = "SELECT site_name FROM $table_name ;";
	$name = $wpdb->get_var( $sql );
	return htmlentities($name);
}

//-------------------------------------------------------------------------------------------------
function rpsl_set_show_login($show_login) {
	global $wpdb;
	$show_login = sanitize_text_field($show_login);
	if($show_login != "Yes" && $show_login != "No" && $show_login != "Click" ) { return false; }
	
	$table_name = $wpdb->prefix . "rpsl_config";
	$update  = $wpdb->prepare("UPDATE " . $table_name . " SET show_login = %s ", $show_login	);
	$updated = $wpdb->query( $update );
	return true;
}

function rpsl_show_login() {
	global $wpdb;
	$table_name = $wpdb->prefix . "rpsl_config";
	$sql = "SELECT show_login FROM $table_name ;";
	$show_login = $wpdb->get_var( $sql );
	return htmlentities($show_login);
}

//*************************************************************************************
// QR Code Construction
//*************************************************************************************
function rpsl_action_site($action) {
	return $action . rpsl_site_code() . rpsl_delim();
}

function rpsl_action_site_name($action) {
	return $action . rpsl_site_code() . rpsl_delim() . rpsl_site_name() . rpsl_delim();
}

//-------------------------------------------------------------------------------------------------
// Generate the QR code data
function rpsl_qr_data($action, $requestid, $username, $anonid = null) {
// Structure is:   vAPP#tV#tV#tV ...
// v=version (currently 'a'), A=action, PP=platform
// fieldcodes are:  n=sitename, r=requestid, u=username, 
//					h= http siteurl, s=https siteurl
	$d = rpsl_delim();
	$qr_data  = "a" . $action . rpsl_site_code();
	$qr_data .= $d . "n" . rpsl_site_name();
	$qr_data .= $d . "r" . $requestid;

	if(!empty($username)) 	{ $qr_data .= $d . "u" . $username; }
	if(!empty($anonid)) 	{ $qr_data .= $d . "a" . $anonid; }

	$url = rpsl_admin_url();
	
	if(strtolower(substr($url,0,8)) == "https://") {
		$qr_data .= $d . "s" . substr($url,8);
	} else {
		$qr_data .= $d . "h" . substr($url,7);
	}
	
	if($action == "S")	{
		$qr_data .= $d . "p" . session_id();
	}
	
	return $qr_data;
}


//-------------------------------------------------------------------------------------------------
// Generate a QR code image from the given data and return the base 64 encoded image
//-------------------------------------------------------------------------------------------------
function rpsl_qr_base64($data, $size = 200) {
	ob_start();
	\PHPQRCode\QRcode::png($data, false, 'L', 4, 2);
	$image_string = base64_encode( ob_get_contents() );
	ob_end_clean();	
	return "data:image/png;base64," . $image_string;
}

//-------------------------------------------------------------------------------------------------
function rpsl_append_admin_url() {
	if (rpsl_site_code() == "W1") { 
		return  (rpsl_delim() . substr(admin_url(), 0, -9)); // Standard location so strip off the "wp-admin/"
	} else {
		return  (rpsl_delim() . admin_url());
	}
}

//-------------------------------------------------------------------------------------------------
function rpsl_admin_url() {
	if (rpsl_site_code() == "W1") { 
		$url = substr(admin_url(), 0, -9); // Standard location so strip off the "wp-admin/"
	} else {
		$url = admin_url();
	}
	if(substr($url, -1) == "/") { $url = substr($url, 0, -1); }
	return $url;
}

//-------------------------------------------------------------------------------------------------
// Returns the URL to the image folder
function rpsl_image_folder_url() {
	$upload = wp_upload_dir();
	return $upload['baseurl'];
}

//-------------------------------------------------------------------------------------------------
// Returns the QR code section delimiter 
function rpsl_delim() {
	return "#";
}

// *******************************************************************************************************
//  Helper Method to instantiate the Rapid Service
// *******************************************************************************************************
function rpsl_get_rapid_service(&$rpsl_message)
{
	$sac_certificate = rpsl_db_get_active_sac_certificate();

	if(empty($sac_certificate)){
		$rpsl_message = "An error occurred, No authentication certificate found to request an identity";
		return;
	}

	$rpsl_server_credentials = array(	
									"cert" 		=> $sac_certificate->public_key,
									"key"  		=> $sac_certificate->private_key,
									"passphrase" 	=> rpsl_keyfile_password(),
									"certfolder"	=> rpsl_cert_folder(),
									"tempfolder"	=> rpsl_temp_folder(),
								);

	$rapidservice = new \Intercede\Rapid($rpsl_server_credentials);

	if (!is_object($rapidservice)) {
		rpsl_trace("Error: Unable to create the RapID Service object");
		$rpsl_message = "An error occurred, please contact your system administrator.";
		return;
	}

	return $rapidservice;
}


//*************************************************************************************
// Given a UUID, Returns the anon id record from the Rapid Service
//*************************************************************************************
function rpsl_get_credential($rpsl_uuid, &$rpsl_request_message) {
	
	// Initialise to blank in case a value was passed in.
	$rpsl_request_message = "";

	// Call to the RapID request function to submit the request and get back the requestID
	$rapid_service = rpsl_get_rapid_service($rpsl_request_message);
	
	if(!empty($rpsl_request_message)) {
		// error occurred in rapid service initialisation
		// We can't recover from this due to mis-configured plugin, persist sessions for now.
		return false;
	}	
	
	try {
		$rpsl_credential_response = $rapid_service->get_credential_by_anonid( $rpsl_uuid );
				
		if(isset($rpsl_credential_response->AnonId)) {
			return $rpsl_credential_response;
		}
		//If message is set, an error has occurred, set up the correct message
		elseif(isset($rpsl_credential_response->Message)){
			$rpsl_request_message = rpsl_extract_rapid_service_message($rpsl_credential_response);
		}
		else
		{
			throw new \Exception("Error: Get Anon ID method returned with no exception, but no message was set");
		}
	} catch (Exception $e) {			
		rpsl_trace("Error: unable to connect to RapID Service - bad RapID site key? \n" . $e->getMessage() );
		$rpsl_request_message = "An error occurred, please contact your system administrator.";
		return false;
	}
}


//*************************************************************************************
// Given a UUID, generate a RapID credential request
//*************************************************************************************
function rpsl_get_new_rapid_request($rpsl_uuid, &$rpsl_request_message) {
	
	// Initialise to blank in case a value was passed in.
	$rpsl_request_message = "";

	// Call to the RapID request function to submit the request and get back the requestID
	$rapid_service = rpsl_get_rapid_service($rpsl_request_message);

	if(!empty($rpsl_request_message)) {
		// error occurred in rapid service initialisation
		return;
	}	
	
	$rpsl_rapid_requestid = null;

	try {
		$rpsl_rapid_response = $rapid_service->request( $rpsl_uuid );

		// If the identifier is set, then we return that as request id
		if(isset($rpsl_rapid_response->Identifier)) {
			$rpsl_rapid_requestid = $rpsl_rapid_response->Identifier;
		}
		// If message is set, an error has occurred, set up the correct message
		elseif(isset($rpsl_rapid_response->Message)){
			$rpsl_request_message = rpsl_extract_rapid_service_message($rpsl_rapid_response);
		}
		else
		{
			throw new \Exception("Error: request method returned with no exception, but no message or identifier was set");
		}
	} catch (Exception $e) {			
		rpsl_trace("Error: unable to connect to RapID Service - bad RapID site key? \n" . $e->getMessage() );
		$rpsl_request_message = "An error occurred, please contact your system administrator.";
	}

	return $rpsl_rapid_requestid;
}

//*************************************************************************************
// Given a Rapid Request Response, Return a normalised error string.
// Must prefix any error message with Error: for some of the other calling code.
//*************************************************************************************
function rpsl_extract_rapid_service_message($rpsl_rapid_response)
{
	$number = $rpsl_rapid_response->Number;
	$message = $rpsl_rapid_response->Message;

	rpsl_trace("Error requesting RapID Id from RapID Service: $number : $message");

	switch($number)
	{
		case 1402:
			return "Error: please contact your system administrator, there are no more remaining licences for new credentials.";
		default:
			return "Error: please contact your system administrator with error code: $number";
		break;
	}
}


//*************************************************************************************
// Check the POST, validate the signed data and return status, UUID and payload
//*************************************************************************************
function rpsl_rapid_verify () {

	global $rpsl_trace_certs;

	// Get some filenames for the signing cert and contents
	$tempfolder = rpsl_temp_folder();
	$pkcs7file  = tempnam($tempfolder, "pk7");
	$certfile   = tempnam($tempfolder, "sig");
	$contfile   = tempnam($tempfolder, "con");
	$issuerfile = tempnam($tempfolder, "iss");
	$untrusted = rpsl_cert_folder() . "/untrusted.pem";        // Any other root cert that won't be matched
	
	try {
	
		$data     = file_get_contents("php://input");         //DIAG	echo "Data is: $data";
		$dataJson = json_decode($data);                       //DIAG	print_r($dataJson);
		$pkcs7    = "";				                          //DIAG	echo "Extracted P7 is: $pkcs7";
		$phone    = "";
		if(property_exists($dataJson, "phone")) {$phone = sanitize_text_field($dataJson->phone);}
		if(property_exists($dataJson, "pkcs7")) {$pkcs7 = $dataJson->pkcs7;}

		// Add the S/MIME Header and save the signed p7 data to file for OpenSSL to read
		$p7handle   = fopen($pkcs7file, "w");
		$mimeheader = "MIME-Version: 1.0\r\n" . "Content-Type: application/pkcs7-mime; smime-type=signed-data;\r\n name=smime.p7m\r\n" . "Content-Transfer-Encoding: base64\r\n\r\n";
		fwrite($p7handle, $mimeheader);
		fwrite($p7handle, $pkcs7);
		fclose($p7handle);

		// turn certificates into a singular pem file
		$issuer_certs = rpsl_db_get_active_issuer_certificates();

		if(count($issuer_certs) <= 0) {
			throw new \Exception("No issuer certs found in the database to be used for verification.");
		}

		$issuer_handle   = fopen($issuerfile, "w");
		foreach($issuer_certs as $cert) {
			fwrite($issuer_handle, $cert->public_key);
			fwrite($issuer_handle, PHP_EOL);
		}	
		fclose($issuer_handle);

		// PKCS7 flags 
		// Do not chain verification of signers certificates: 
		// that is don't use the certificates in the signed message as untrusted CAs.
		$flags = PKCS7_NOCHAIN;
		
		// Now verify and parse the data, If the json pkcs7 written to the file is not a pkcs7 data packet
		// the openssl libary will fail to parse and verify it. If it is a valid packet but signed with
		// incorrect certificates openssl will also fail to parse and verify it.
		// pkcs7 file is the packet to verify 
		// certfile is the certificate used to sign the p7
		// caroot cert is the trusted ca to use in the verification process
		// untrusted is a blank pem 
		// contfile is the filepath which contains the verified content of the p7
		$result = openssl_pkcs7_verify($pkcs7file, $flags, $certfile, array($issuerfile), $untrusted , $contfile);
		
		$content_handle = fopen($contfile, "r");
		$data = fread($content_handle, filesize($contfile));
		fclose($content_handle);	
		rpsl_log_openssl_error();
		
		$signcert = file_get_contents($certfile);
		$certinfo = openssl_x509_parse($signcert);
		$uuid     = substr($certinfo["name"],4);   // Strip off the "/"

		rpsl_set_device_last_used($uuid);
		
		//DIAG 
		$certAsText = print_r($certinfo, true);
		if ($rpsl_trace_certs) rpsl_trace("Certificate: \n $certAsText"); 
		if ($rpsl_trace_certs) rpsl_trace("Raw cert is: \n $signcert"); 
		rpsl_trace("Data: $data");
		
		// Clean up the temporary files
		clean_up_temp_file($certfile);
		clean_up_temp_file($contfile);
		clean_up_temp_file($pkcs7file);
		clean_up_temp_file($issuerfile);
		
		// If all is ok, pass back raw post data in result array for calling method to use as they see fit.
		// otherwise return an error array.
		$result = array (
					'status'    => 0,
					'message'   => 'ok',
					'uuid'      => $uuid,
					'data'      => $data,
					'phone'     => $phone,
						);
	} catch(Exception $e) {
	
		rpsl_trace("Verify Error: " . $e->getMessage());
		
		// Clean up the temporary files
		clean_up_temp_file($certfile);
		clean_up_temp_file($contfile);
		clean_up_temp_file($pkcs7file);
		clean_up_temp_file($issuerfile);

		$result = array (
					'status'    => 100,
					'message'   => 'The RapID trusted issuing CA certificate is missing. Please check the WordPress RapID Settings',
					'uuid'      => '',
					'data'      => '',
					'phone'     => '',
						);
	}

	return $result;
}

function clean_up_temp_file($tempfile_path) {
	
	if(!empty($tempfile_path) && file_exists($tempfile_path)) {
		unlink($tempfile_path);
	}
}

/*************************************************************
* function to create a token file to trigger the ajax polling.
*************************************************************/
function rpsl_create_session_token_file($session, $content){
		
		//Write a temporary file for the polling script.
		$tempfolder = rpsl_temp_folder();
		rpsl_delete_old_files($tempfolder);  // Housekeeping
		
		$file = $tempfolder . "/" . $session;		
		rpsl_trace("File to be written: $file");
		
		$fp = fopen($file, "w");
		fwrite($fp, $content);
		fclose($fp);
}

function rpsl_cleanup_expired_session_rows() {

	global $wpdb;
	$table_name = $wpdb->prefix . "rpsl_sessions";

	// Delete existing entries in the session table for this session 
	$cleanup = "SELECT * FROM " . $table_name . " WHERE requested < (UTC_TIMESTAMP() - INTERVAL 20 MINUTE)";
	$expired = $wpdb->get_results( $cleanup );
	// For each of the expired actions, we need to work out whether there is an orphaned credential
	// That we need to query for.
	
	$rows_to_delete = array();
	foreach($expired as $expired_row) {
		if(	!isset($expired_row->uuid) || 
			!isset($expired_row->loginname) ||
			(isset($expired_row->action) && $expired_row->action == "login")) {
			// its a login action ignore
			// no login name set on row, can't do anything, ignore.
			// it doesn't have a uuid row set, can't do anything anyway, ignore.
			$rows_to_delete[] = $expired_row->rpsession; // This adds it to the end of the array
			continue;
		}

		// we have an expired_row with a uuid, is that already matched to a user
		$already_linked = rpsl_get_user_by_uuid($expired_row->uuid);

		if($already_linked != null) {
			$rows_to_delete[] = $expired_row->rpsession; // This adds it to the end of the array
			continue;
		}

		// make call to Rapid Service to confirm if the credential was collected
		// before we remove the session row.
		$error_status = "";
		$credential = rpsl_get_credential($expired_row->uuid, $error_status);
		if($credential === false) {
			rpsl_diag("The call to the RapID Service returned false, this means that the service was not contactable, end the cleanup for a future version.");
			// Something went wrong with the rapid service, we need to retry next time.
			// Don't keep looping as we know it will continue to fail, so we just return
			// waiting for the next clean up call.
			return;
			// Extend time limit by 20 minutes.
			// $wpdb->query($wpdb->prepare("UPDATE " . $table_name . " SET requested = DATE_ADD(NOW(), INTERVAL 20 MINUTES) WHERE rpsession = %s)", $expired_row->rpsession));
		}

		// At this point, either the credential was collected, the login wasn't 
		// A user, or the credential wasn't collected and so is expired, 
		// or there was an error on getting the credential
		// so we delete it
		
		$rows_to_delete[] = $expired_row->rpsession; // This adds it to the end of the array

		if(!empty($error_status)) {
			// something went wrong which we can't recover from,  
			// e.g. anon id requested, service does not know about.
			// Either it's expired and been cle'
			// service was contactable but can't give us a record to process on
			continue;
		}

		if(isset($credential->Status) && $credential->Status == "Collected") {
			// This credential was collected, lets link it up
			// Add new device based on getting the user from the login name in the session.
			$user = get_user_by('login', $expired_row->loginname);

			if(is_a($user, 'WP_User')){
				// User exists, add as a device.
				// Default device name of mobile device, will be updated when they log in.
				rpsl_add_device( $user->ID, $expired_row->uuid, "Mobile Device");
			}
		}
	}

	// Have now been through all the rows in the collection. Clean up where required.
	// Have to implode with quotes and comma to create correct sql in clause structure
	$rpsessions = implode("','", $rows_to_delete);
	$cleanup_query = "DELETE FROM " . $table_name . " WHERE rpsession in ('$rpsessions')";
	rpsl_diag("Deleting Rows: $cleanup_query");
	$wpdb->query($cleanup_query);
}

/*************************************************************
 * Adding Scripts
/*************************************************************/

function rpsl_adding_scripts() {
	wp_register_script('rpsl_configure_wordpress_registration', plugins_url('scripts/rpsl_configure_wordpress_registration.js', __FILE__), array('jquery'),'1.9', true);
}

/*************************************************************
 * rpsl_UUID class
 * The following class generates VALID RFC 4122 COMPLIANT
 * Version 4 Universally Unique IDentifiers (UUID)
 *************************************************************/
class rpsl_UUID
{
	/*********************************************************
	 * Version 4 UUIDs are pseudo-random.
	 *********************************************************/
	public static function v4() {
		return sprintf('%04x%04x-%04x-%04x-%04x-%04x%04x%04x',
		mt_rand(0, 0xffff), mt_rand(0, 0xffff), // 32 bits for "time_low"
		mt_rand(0, 0xffff),                     // 16 bits for "time_mid"
		mt_rand(0, 0x0fff) | 0x4000,			// 16 bits for "time_hi_and_version",
												// Four most significant bits hold version number 4
		mt_rand(0, 0x3fff) | 0x8000,    		// 16 bits, 8 bits for "clk_seq_hi_res",
												// 8 bits for "clk_seq_low",
												// Two most significant bits hold zero and one for variant DCE1.1
		mt_rand(0, 0xffff), mt_rand(0, 0xffff), mt_rand(0, 0xffff)	// 48 bits for "node"
		);
	}
}